//
// EregionAnalyse.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to calculate the optimum radius and centroid
//              for a point source extraction
//
#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "CalServer.h"
#include "Dal.h"

#include "EregionAnalyse.h"
#include "ImageRegion.h"
#include "CalSetter.h"

/* Method: EregionAnalyse() - Constructor */
EregionAnalyse::EregionAnalyse():
srcreg(0),
bckreg(0),
_calset(0)
{
  PROC;

  /* Get the parameters from this command line */
    string imagefile = stringParameter("imageset");
    string bckgndfile = stringParameter("bkgimageset");
    string exposurefile = stringParameter("exposuremap");
    string srcexp = stringParameter("srcexp");
    string backexp = stringParameter("backexp");

  /* Attempt to set the CAL state using the input image header */
    _calset = new CalSetter();
    _calset->fromImage(imagefile);
    _calset->setAccuracyLevel(EXTENDED);

  /* We can't currently optimise timing and burst mode data, so
     set status bad and exit early in these cases */
    Xmm::Mode windowMode = calServer.state()->mode()->get();

    if (windowMode == Xmm::FAST_TIMING ||
        windowMode == Xmm::FAST_BURST ||
        windowMode == Xmm::FAST_UNCOMPRESSED || 
        windowMode == Xmm::FAST_COMPRESSED) {
       
       setParameter("status",false);

       errstr<<" Timing/Burst mode - setting status false"
          << Sas::message(Msg::App, Msg::Verbose);

       ALTRETURN;
    }

  /* Set the instrument as a local variable */
    //Xmm::Instrument instr = calServer.instrument()->id();

  /* Set up an ImageRegion object for the source area */
    srcreg = new ImageRegion(imagefile, srcexp, "filtpixlist_src.ds:");

    errstr<<" Set up source image region"
          << Sas::message(Msg::App, Msg::Verbose);

  /* Check out a few methods */
    double delrad = srcreg->getRadius();
    double deltheta = srcreg->theta();
    double delcnts = srcreg->counts();
    double delflux = srcreg->flux();

    errstr<<" Radius: "<<delrad
          <<" Theta: "<<deltheta
          <<" Phi: "<<srcreg->phi() << endl
          <<" Counts: "<<delcnts << endl
          <<" numpix: "<<srcreg->number_of_pixels() << endl
          <<" flux: "<<delflux << endl
          <<" pixelArea: "<<srcreg->pixelArea()
          <<" xwid: "<<srcreg->xWidth()
          <<" ywid: "<<srcreg->yWidth()
          << Sas::message(Msg::App, Msg::Verbose);

  /* Process the background information: If a background map has been 
     supplied use this in conjunciton with the source spatial expression,
     If a map hasn´t been set and a background spatial expression has
     been supplied find the background flux per unit area from the source
     image. Otherwise use the value supplied in the parameter backval */

    bool backgnd_region_set=true;

    if (bckgndfile != "NotSet")
       bckreg = new ImageRegion(bckgndfile, srcexp, "filtpixlist_bgd.ds:");
    else if (backexp != "NotSet")
    {
       bckreg = new ImageRegion(imagefile, backexp, "filtpixlist_bgd.ds:");
    } 
    else {
       bckreg=0;
       backgnd_region_set=false;
    }

   /* Find the background per unit area */
    double bck_flux=0;
    if (backgnd_region_set) {

       bck_flux = bckreg->flux();

       errstr<<" Backgnd: Radius: "<<bckreg->getRadius()
          <<" Theta: "<<bckreg->theta()
          <<" Counts: "<<bckreg->counts()<< endl
          <<" numpix: "<<bckreg->number_of_pixels() << endl
          <<" pixelArea: "<<bckreg->pixelArea()
          <<" xwid: "<<bckreg->xWidth()
          <<" ywid: "<<bckreg->yWidth()
          <<" Background flux: "<<bck_flux
          << Sas::message(Msg::App, Msg::Verbose);

    } else {  // get backgnd flux from command line
       bck_flux = realParameter("backval");
    }

  /* Centroid the source */
    CoordPair centre = srcreg->centroid(5);

  /* Find the number of counts in the source area */
    double src_cnts = srcreg->counts();

  /* Should get energy from selection expression but lets approximate
     to 1.5 keV for now */
    double energy=1500.;

  /* Get the off-axis angle and azimuthal angle of the source box */
    double theta = srcreg->theta();
    double phi = srcreg->phi();
    double src_rad = srcreg->getRadius();

  /* Find the optimum radius (in arcsecs) of the source box */
    double optrad = optimumRadius(src_cnts, src_rad, bck_flux,
                                  energy, theta, phi);

  /* Convert optimum radius to image units */
    double optrad_image = optrad * srcreg->asecToImgUnits();

  /* Find the exposure time for the source. If an exposure map has been
     provided use this, otherwise just use the exposure keyword from 
     the source image which is only correct if the source is on-axis */
    double exposure = 0.0;
    if (exposurefile != "NotSet") {
       expreg = new ImageRegion(exposurefile, srcexp, "filtpixlist_exp.ds:");
       exposure = expreg->centralPixelValue();
    } else {
       exposure = srcreg->exposure();
    }

  /* Calculate the background subtracted, EEF corrected, source count rate */
    calcSrcCountRate(srcreg, bckreg, bck_flux, backgnd_region_set,
                                  energy, exposure);

  /* Get the initial source region centre in RA/DEC */
    CoordPair initcentre = srcreg->celestialCentre();


    /* 
       AI:SOC-SPR-2687. 
       For some reason in Mac OS X you cannot call a method inside a 
       errstr statement. So we have defined some variables where we store the 
       values and the we print them out using the errstr and the SAS log infrastucutre.
     */

    double x_coord = initcentre.x();
    double y_coord = initcentre.y();
    double cnts_per_pixel = srcreg->countsperpix();
    double x_centre=centre.x();
    double y_centre=centre.y();
    double encircle_fac=_source.ee;
    double bkg_sub_count = _source.bkgsub_counts;
    double bkg_sub_err =_source.bkgsub_counts_error;
    double bkg_sub_cr = _source.bkgsub_count_rate;
    double bkg_sub_cr_err = _source.bkgsub_count_rate_error;

    errstr<<" input region centre: "<<x_coord <<" "<<y_coord <<endl
	  <<" counts in source region: "<<src_cnts <<endl
	  <<" src region cnts per pixel: " <<cnts_per_pixel <<endl
	  <<" exposure time: "<<exposure<<endl
	  <<" xcentroid: "<<x_centre <<endl
	  <<" ycentroid: "<< y_centre <<endl
	  <<" optradius: "<<optrad << " arcsecs "<<optrad_image << " image units" <<endl
	  <<" encircled energy factor: "<<  encircle_fac <<endl
	  <<" Bckgnd subtracted source cnts: "<< bkg_sub_count <<" +/- "<< bkg_sub_err <<endl
	  <<" Bckgnd subtracted source c/r: "<< bkg_sub_cr <<" +/- "<< bkg_sub_cr_err << Sas::message(Msg::App, Msg::Always);

  /* Set output parameters */
    setParameter("xcentroid",centre.x());
    setParameter("ycentroid",centre.y());
    setParameter("optradius",optrad_image);
    setParameter("srccnts",src_cnts);
    setParameter("status",true);

  ENDPROC;
}

/* Destructor */
EregionAnalyse:: ~EregionAnalyse()
{
  PROC;

    delete srcreg;            srcreg=0;
    delete bckreg;            bckreg=0;
    delete _calset;           _calset=0;

  ENDPROC;
};

/* Method: optimumRadius - finds the optimum radius for source extraction
        from the source and background counts */
double EregionAnalyse::optimumRadius(double src_cnts, double sboxrad, 
                     double bck_flux,
                     double energy, double theta, 
                     double phi)
{
  /* Set up the encircled energy function */
   PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);
   Psf & p = psfs->psf(energy, theta, phi);
   _psf = &p;

   errstr<<"EEF set up with Energy: "<<energy<<" Theta: "<<theta
         <<" Phi: "<<phi
            << Sas::message(Msg::App, Msg::Verbose);

  /* Get the total number of counts in source */
   /* Find the fraction of counts in the source box radius */
   double asec2mm = 27.5;
   double test_rad = sboxrad / asec2mm;
   double ee = _psf->encircledEnergy(test_rad); 

   /* Find the background counts in the source box */
   double bck_cnts = bck_flux * M_PI * sboxrad * sboxrad;

   errstr<<" Back: "<<bck_cnts<<" "<<bck_flux<<" "<<sboxrad
         <<" EE(2.4) "<<_psf->encircledEnergy(2.4)
            << Sas::message(Msg::App, Msg::Verbose);


   double totsrc = (src_cnts - bck_cnts) / ee;

   errstr<<" SRC: "<<totsrc<<" "<<src_cnts<<" "<<bck_cnts
            << Sas::message(Msg::App, Msg::Verbose);

   double snr=0, snrmax=0, maxrad=0;

  /* Loop over test radii between 0.0 and 5.0 arcminutes in one arcsec bins */

   for (int i=1; i<301; i++)
   {
    /* Get the fraction of source counts within a circle of this radius.
       NB: This takes forever to complete unless the PSF mode is EXTENDED
       so use a mickey-mouse formulation for other modes */

      if (calServer.state()->accuracyLevel()->get() == EXTENDED) {
         test_rad = i / asec2mm;  // test radius is in mm (CamCoords2) 
         ee = _psf->encircledEnergy(test_rad);
       }
      else {
         ee = i / 300.0;
       }

    /* Calculate the S/N */
      snr = totsrc * ee / sqrt (totsrc * ee + bck_flux * M_PI
                                   * i * i);
      errstr<<" Optimum try: "<<i<<" "<<totsrc<<" "<<ee<<" "
            <<bck_flux<<" "<<M_TWO_PI<<" "<<snr
            << Sas::message(Msg::App, Msg::Noisy);

    /* Update the maximum */
      if (snr > snrmax) 
      {
         snrmax = snr;
         maxrad = i;
      }
   }

  /* Set artifical mimimum extent if optimum radius too small */
   maxrad = maxrad < 10 ? 10 : maxrad;
 
   errstr<<" Optimum radius is: "<<maxrad
   << Sas::message(Msg::App, Msg::Verbose);

   RETURN maxrad;
}

/* Method: calcSrcCountRate - returns the background subtracted source
                               count rate * EEF */
void EregionAnalyse::calcSrcCountRate(ImageRegion * srcreg,
                     ImageRegion * bckreg,
                     double bck_flux,
                     bool backgnd_region_set,
                     double energy, double exposure) 
{
  /* Set up the encircled energy function */
   PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);
   Psf & p = psfs->psf(energy, srcreg->theta(), srcreg->phi());
   _psf = &p;

  /* Get the total number of counts in source */
   /* Find the fraction of counts in the source box radius */
   double asec2mm = 27.5;
   double test_rad = srcreg->getRadius() / asec2mm;
   double ee = _psf->encircledEnergy(test_rad); 

   double src_bck_cnts=0, totsrc=0, toterror=0;

   /* Find the background counts in the source box  and then the
      background subtracted source count rate and error */
   if (backgnd_region_set)
   {
      double area_ratio = srcreg->area() / bckreg->area();
      double src_bck_cnts = bckreg->counts() * area_ratio;

      errstr<<" Back: "<<src_bck_cnts<<" "<<area_ratio
         <<" EE(2.4) "<<_psf->encircledEnergy(2.4)
            << Sas::message(Msg::App, Msg::Verbose);

     /* Calc bckgnd subtracted counts corrected for PSF */
      totsrc = (srcreg->counts() - src_bck_cnts) / ee;
      toterror = sqrt(srcreg->counts() + bckreg->counts() 
                                       * area_ratio*area_ratio) / ee;

   } else { // Using entered bckgnd flux
      src_bck_cnts = bck_flux * srcreg->area();
      totsrc = (srcreg->counts() - src_bck_cnts) / ee;
      toterror = 0.0;  // set to zero in this case
   }

 /* Set the background subtracted source count rate and error */
   _source.bkgsub_counts = totsrc;
   _source.bkgsub_counts_error = toterror;
   _source.bkgsub_count_rate = totsrc / exposure;
   _source.bkgsub_count_rate_error = toterror / exposure;
   _source.ee = ee;

}

   
